@@ -39,16 +39,16 @@ from requests_oauthlib import OAuth2Session
 from . import status
 from .exceptions import (
     DynamicsException,
-    DuplicateRecordError,
-    PayloadTooLarge,
-    APILimitsExceeded,
-    OperationNotImplemented,
-    WebAPIUnavailable,
     ParseError,
     AuthenticationFailed,
     PermissionDenied,
     NotFound,
     MethodNotAllowed,
+    DuplicateRecordError,
+    PayloadTooLarge,
+    APILimitsExceeded,
+    OperationNotImplemented,
+    WebAPIUnavailable,
 )
 
 
@@ -60,49 +60,56 @@ __all__ = [
 logger = logging.getLogger(__name__)
 
 method_type = Literal["get", "post", "patch", "delete"]
-orderby_type = Dict[str, Literal["desc", "asc"]]
+orderby_type = Dict[str, Literal["asc", "desc"]]
+filter_type = Union[Set[str], List[str]]
 
-# $select takes List[str], $fitler takes List[str] or Set[str],
-# $top takes int, and $orderby takes orderby_type.
-# Nested $expand takes a dict like defined here
 expand_keys = Literal["select", "filter", "top", "orderby", "expand"]
 expand_values = Union[List[str], Set[str], int, orderby_type, Dict]
-expand_subcommands = Dict[expand_keys, expand_values]
+expand_type = Dict[str, Dict[expand_keys, expand_values]]
+
+
+class _sentinel:
+    """Sentinel value."""
+
+    def __bool__(self):
+        return False
 
 
 class DynamicsClient:
     """Dynamics client for making queries from a Microsoft Dynamics 365 database."""
 
-    request_counter = 0
+    request_counter: int = 0
 
     def __init__(self, api_url: str, token_url: str, client_id: str, client_secret: str, scope: List[str]):
-        """Establish a Microsoft Dynamics 365 Dataverse API client connection.
+        """Establish a Microsoft Dynamics 365 Dataverse API client connection
+        using OAuth 2.0 Client Credentials Flow. Client Credentials require an application user to be
+        created in Dynamics, and granting it an appropriate security role.
 
-        :param api_url: Url in form: 'https://[Organization URI]/api/data/v{api_version}'
-        :param token_url: Url in form: 'https://[Dynamics Token URI]/path/to/token'
-        :param client_id: Client id (e.g. UUID).
-        :param client_secret: Client secret (e.g. OAuth password).
-        :param scope: List of urls in form: 'https://[Organization URI]/scope'.
-                      Defines the database records that the API connection has access to.
+        :param api_url: API root URL. Format: https://[Organization URI]/api/data/v{api_version}
+        :param token_url: URL to the Dynamics/Azure token endpoint.
+                          Format: https://[Dynamics Token URI]/path/to/token
+        :param client_id: Dynamics User ID.
+        :param client_secret: Dynamics User Secret that proves its identity when password is required.
+        :param scope: List of urls that define the database records that the API connection has access to.
+                      Most likely in this format: https://[Organization URI]/.default
         """
 
         # [sic] Assure that url ends in forward slash
-        self.api_url = api_url.rstrip("/") + "/"
+        self._api_url = api_url.rstrip("/") + "/"
 
         # [sic] Scope not given here, since we don't have a token yet
-        client = BackendApplicationClient(client_id=client_id)
-        self.api = OAuth2Session(client=client)
+        self._api = OAuth2Session(client=BackendApplicationClient(client_id=client_id))
 
         # Fetch token
-        self.api.fetch_token(token_url=token_url, client_id=client_id, client_secret=client_secret, scope=scope)
+        self._api.fetch_token(token_url=token_url, client_id=client_id, client_secret=client_secret, scope=scope)
 
         # Query options
-        self._select = ""
-        self._expand = ""
-        self._top = ""
-        self._filter = ""
-        self._orderby = ""
-        self._count = ""
+        self._select: List[str] = []
+        self._expand: expand_type = {}
+        self._filter: filter_type = []
+        self._orderby: orderby_type = {}
+        self._top: int = 0
+        self._count: bool = False
 
         self._table = ""
         self._action = ""
@@ -110,7 +117,14 @@ class DynamicsClient:
         self._add_ref_to_property = ""
         self._pre_expand = ""
 
-        self.headers = {}
+        self._headers: Dict[str, str] = {}
+        self._pagesize: int = 1000
+
+    def __getitem__(self, key):
+        return self.headers[key]
+
+    def __setitem__(self, key, value):
+        self.headers[key] = value
 
     @classmethod
     def from_environment(cls):
@@ -133,53 +147,77 @@ class DynamicsClient:
 
         return cls(base_url, token_url, client_id, client_secret, scope)
 
+    def fetch_schema(self, to_file: bool = False) -> Optional[str]:
+        """Fetch Dynamics schema for observation."""
+
+        import xml.etree.ElementTree as ET
+        from operator import attrgetter
+        from xml.dom.minidom import parseString as pretty_xml
+
+        def sortchildrenby(parent, attr):
+            parent[:] = sorted(parent, key=lambda child: child.get(attr, ""))
+            parent[:] = sorted(parent, key=attrgetter("tag"))
+
+        ET.register_namespace("edmx", "http://docs.oasis-open.org/odata/ns/edmx")
+        ET.register_namespace("", "http://docs.oasis-open.org/odata/ns/edm")
+
+        logger.info("Fetching Schema.")
+        request = self._api.get(self._api_url + "$metadata")
+        logger.info("Schema fetched. Formatting...")
+        xml_string = pretty_xml(request.text).toprettyxml(indent="  ")
+
+        tree = ET.ElementTree(ET.fromstring(xml_string))
+        root = (
+            tree.getroot()
+                .find("edmx:DataServices", namespaces={"edmx": "http://docs.oasis-open.org/odata/ns/edmx"})
+                .find("Schema", namespaces={"": "http://docs.oasis-open.org/odata/ns/edm"})
+        )
+
+        sortchildrenby(root, "Name")
+        for child in root:
+            sortchildrenby(child, "Name")
+
+        xml_prettyfied = pretty_xml(ET.tostring(root)).toprettyxml(indent="  ")
+        xml_prettyfied = "\n".join([e for e in xml_prettyfied.split("\n") if e.strip()])
+
+        if not to_file:
+            return xml_prettyfied
+
+        with open(f"dynamics_schema.xml", "w+") as f:
+            f.write(xml_prettyfied)
+
     @property
     def current_query(self) -> str:
         """Constructs query from current options, leaving out empty ones."""
 
-        query = self.api_url + self.table
+        query = self._api_url + self.table
 
         if self.row_id:
             query += f"({self.row_id})"
         if self.pre_expand:
             query += f"/{self.pre_expand}"
         if self.action:
+            if query[-1] != "/":
+                query += "/"
             query += self.action
         if self.add_ref_to_property and not self.pre_expand and not self.action:
             query += f"/{self.add_ref_to_property}/$ref"
-        if (qo := self.compile_query_options()) and not self.add_ref_to_property:
+        if (qo := self._compile_query_options()) and not self.add_ref_to_property:
             query += qo
 
         return query
 
-    def reset_query(self):
-        """Resets the query options and table selection."""
-        self._select = ""
-        self._expand = ""
-        self._top = ""
-        self._filter = ""
-        self._orderby = ""
-        self._count = ""
-
-        self._table = ""
-        self._action = ""
-        self._row_id = ""
-        self._add_ref_to_property = ""
-        self._pre_expand = ""
-
-        self.headers = {}
-
-    def compile_query_options(self) -> str:
+    def _compile_query_options(self) -> str:
         query_options = "&".join(
             [
                 statement
                 for statement in [
-                    self._expand,
-                    self._select,
-                    self._filter,
-                    self._top,
-                    self._count,
-                    self._orderby,
+                    self._compile_expand(),
+                    self._compile_select(),
+                    self._compile_filter(),
+                    self._compile_top(),
+                    self._compile_count(),
+                    self._compile_orderby(),
                 ]
                 if statement
             ]
@@ -187,52 +225,75 @@ class DynamicsClient:
 
         return f"?{query_options}" if query_options else ""
 
-    def set_default_headers(self, operation: method_type):
+    @property
+    def headers(self) -> Dict[str, str]:
+        """HTTP request headers."""
+        return self._headers
+
+    def reset_query(self):
+        """Resets all client options and headers."""
+        self._select: List[str] = []
+        self._expand: expand_type = {}
+        self._filter: filter_type = []
+        self._orderby: orderby_type = {}
+        self._top: int = 0
+        self._count: bool = False
+
+        self._table = ""
+        self._action = ""
+        self._row_id = ""
+        self._add_ref_to_property = ""
+        self._pre_expand = ""
+
+        self._headers: Dict[str, str] = {}
+
+    def set_default_headers(self, method: method_type):
+        """Sets per method default headers. Does not override user added headers."""
+
         self.headers.setdefault("OData-MaxVersion", "4.0")
         self.headers.setdefault("OData-Version", "4.0")
         self.headers.setdefault("Accept", "application/json; odata.metadata=minimal")
 
-        if operation in ["post", "patch", "delete"]:
+        if method in ["post", "patch", "delete"]:
             self.headers.setdefault("Content-Type", "application/json; charset=utf-8")
 
-        if operation in ["post", "patch"]:
+        if method in ["post", "patch"]:
             self.headers.setdefault("Prefer", "return=representation")
             self.headers.setdefault("MSCRM.SuppressDuplicateDetection", "false")
 
-        if operation in ["patch"]:
+        if method in ["patch"]:
             self.headers.setdefault("If-None-Match", "null")
             self.headers.setdefault("If-Match", "*")
 
-        self.headers.setdefault("Prefer", "odata.maxpagesize=1000")
+        self.headers.setdefault("Prefer", f"odata.maxpagesize={self.pagesize}")
 
-    @staticmethod
-    def error_handling(status_code: int, error_message: str, method: method_type):
+    def _error_handling(self, status_code: int, error_message: str, method: method_type):
         """Error handling based on these expected error statuses:
         https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/compose-http-requests-handle-errors#identify-status-codes
         """
 
         if status_code == status.HTTP_400_BAD_REQUEST:
-            raise ParseError(message=error_message)
+            raise ParseError(detail=error_message)
         elif status_code == status.HTTP_401_UNAUTHORIZED:
-            raise AuthenticationFailed(message=error_message)
+            raise AuthenticationFailed(detail=error_message)
         elif status_code == status.HTTP_403_FORBIDDEN:
-            raise PermissionDenied(message=error_message)
+            raise PermissionDenied(detail=error_message)
         elif status_code == status.HTTP_404_NOT_FOUND:
-            raise NotFound(message=error_message)
+            raise NotFound(detail=error_message)
         elif status_code == status.HTTP_405_METHOD_NOT_ALLOWED:
-            raise MethodNotAllowed(method=method, message=error_message)
+            raise MethodNotAllowed(method=method, detail=error_message)
         elif status_code == status.HTTP_412_PRECONDITION_FAILED:
-            raise DuplicateRecordError(message=error_message)
+            raise DuplicateRecordError(detail=error_message)
         elif status_code == status.HTTP_413_REQUEST_ENTITY_TOO_LARGE:
-            raise PayloadTooLarge(message=error_message)
+            raise PayloadTooLarge(detail=error_message)
         elif status_code == status.HTTP_429_TOO_MANY_REQUESTS:
-            raise APILimitsExceeded(message=error_message)
+            raise APILimitsExceeded(detail=error_message)
         elif status_code == status.HTTP_501_NOT_IMPLEMENTED:
-            raise OperationNotImplemented(message=error_message)
+            raise OperationNotImplemented(detail=error_message)
         elif status_code == status.HTTP_503_SERVICE_UNAVAILABLE:
-            raise WebAPIUnavailable(message=error_message)
+            raise WebAPIUnavailable(detail=error_message)
         else:
-            raise DynamicsException(message=error_message)
+            raise DynamicsException(detail=error_message)
 
     def GET(self, not_found_ok: bool = False, next_link: Optional[str] = None) -> List[Dict[str, Any]]:
         """Make a request to the Dataverse API with currently added query options.
@@ -245,9 +306,9 @@ class DynamicsClient:
         self.set_default_headers("get")
 
         if next_link is not None:
-            response = self.api.get(next_link, headers=self.headers)
+            response = self._api.get(next_link, headers=self.headers)
         else:
-            response = self.api.get(self.current_query, headers=self.headers)
+            response = self._api.get(self.current_query, headers=self.headers)
 
         data = response.json()
 
@@ -257,12 +318,12 @@ class DynamicsClient:
         count = data.get("@odata.count", "")
 
         if errors:
-            self.error_handling(response.status_code, errors.get("message"), method="get")
+            self._error_handling(response.status_code, errors.get("message"), method="get")
         elif not entities:
             if not_found_ok:
                 return []
             message = "No records matching the given criteria."
-            self.error_handling(status.HTTP_404_NOT_FOUND, message, method="get")
+            self._error_handling(status.HTTP_404_NOT_FOUND, message, method="get")
 
         # Fetch more data if needed
         for i, row in enumerate(entities):
@@ -274,7 +335,7 @@ class DynamicsClient:
                     # Therefore, if the @odata.next link appears before that, we can ignore it.
                     #
                     key = column_key[:-15]
-                    if len(entities[i][key]) < 1000:
+                    if len(entities[i][key]) < self.pagesize:
                         row.pop(column_key)
                         continue
 
@@ -307,7 +368,7 @@ class DynamicsClient:
 
         # [sic] POSTing data in dict form doesn't work for some reason...
         data = json.dumps(data).encode()
-        response = self.api.post(self.current_query, data=data, headers=self.headers)
+        response = self._api.post(self.current_query, data=data, headers=self.headers)
 
         if response.status_code == status.HTTP_204_NO_CONTENT:
             return {}
@@ -315,9 +376,7 @@ class DynamicsClient:
         data = response.json()
 
         if errors := data.get("error"):
-            if errors["code"] == "0x80040265":
-                raise OverlappingBooking()  # No logging since this likely happens a lot
-            self.error_handling(response.status_code, errors["message"], method="post")
+            self._error_handling(response.status_code, errors["message"], method="post")
 
         return data
 
@@ -332,7 +391,7 @@ class DynamicsClient:
         self.set_default_headers("patch")
 
         data = json.dumps(data).encode()
-        response = self.api.patch(self.current_query, data=data, headers=self.headers)
+        response = self._api.patch(self.current_query, data=data, headers=self.headers)
 
         if response.status_code == status.HTTP_204_NO_CONTENT:
             return {}
@@ -340,19 +399,17 @@ class DynamicsClient:
         data = response.json()
 
         if errors := data.get("error"):
-            if errors["code"] == "0x80040265":
-                raise OverlappingBooking()  # No logging since this likely happens a lot
-            self.error_handling(response.status_code, errors["message"], method="patch")
+            self._error_handling(response.status_code, errors["message"], method="patch")
 
         return data
 
-    def DELETE(self):
+    def DELETE(self) -> Dict[str, Any]:
         """Delete row in a table. Must have 'table' and 'row_id' attributes set."""
 
         self.request_counter += 1
         self.set_default_headers("delete")
 
-        response = self.api.delete(self.current_query, headers=self.headers)
+        response = self._api.delete(self.current_query, headers=self.headers)
 
         if response.status_code == status.HTTP_204_NO_CONTENT:
             return {}
@@ -360,7 +417,9 @@ class DynamicsClient:
         data = response.json()
 
         if errors := data.get("error"):
-            self.error_handling(response.status_code, errors["message"], method="delete")
+            self._error_handling(response.status_code, errors["message"], method="delete")
+
+        return data
 
     @property
     def table(self) -> str:
@@ -369,12 +428,25 @@ class DynamicsClient:
 
     @table.setter
     def table(self, value: str):
-        self._table =
+        self._table = value
 
     @property
     def action(self) -> str:
-        """API action to take.
-        https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/use-web-api-actions
+        """Set the Dynamics Web API action or function to use.
+
+        It is recommended to read the API Function Reference:
+        https://docs.microsoft.com/en-us/dynamics365/customer-engagement/web-api/functions
+
+        ...and how to Use Web API Functions:
+        https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/use-web-api-functions
+
+        ...as well as the API Action Reference
+        https://docs.microsoft.com/en-us/dynamics365/customer-engagement/web-api/actions
+
+        ...and how to Use Web API Actions:
+        https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/use-web-api-actions.
+
+        You can input the action/function as a string, or use the included `act` and `fnc` objects construct it.
         """
         return self._action
 
@@ -388,7 +460,7 @@ class DynamicsClient:
         If the table has an alternate key defined, you can use
         'foo=bar' or 'foo=bar,fizz=buzz' to retrive a single row:
         https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/retrieve-entity-using-web-api#retrieve-using-an-alternate-key
-        Alternate keys are not on by default in Dynamics, so those might not work at all.
+        Alternate keys are not enabled by default in Dynamics, so those might not work at all.
         """
         return self._row_id
 
@@ -399,7 +471,7 @@ class DynamicsClient:
     @property
     def add_ref_to_property(self) -> str:
         """Add reference for this navigation property. This indicates,
-        that POST data will contain the API url to a matching rowid
+        that POST data will contain the API url to a matching row id
         in the table this navigation property is meant to link to,
         like this: "@odata.id": "<API URI>/<table>(<id>)".
 
@@ -417,7 +489,7 @@ class DynamicsClient:
         self._add_ref_to_property = value
 
     @property
-    def pre_expand(self):
+    def pre_expand(self) -> str:
         """Expand/navigate to some linked table in this table
         before taking any queryoptions into account.
         This will save you having to use the expand statement itself,
@@ -445,72 +517,32 @@ class DynamicsClient:
             self.headers.pop("Prefer")
 
     @property
-    def select(self) -> str:
+    def select(self) -> List[str]:
         """Get current $select statement"""
         return self._select
 
     @select.setter
     def select(self, items: List[str]):
-        """Set $select statement. Limits the properties returned from the current entity.
-        Use expand to limit properties in related items
-        """
-        self._select = "$select=" + ",".join([key for key in items])
+        """Set $select statement. Select which columns are returned from the table."""
+        self._select = items
+
+    def _compile_select(self, values: List[str] = _sentinel) -> str:
+        if values is _sentinel:
+            values = self._select
+
+        return "$select=" + ",".join([key for key in values]) if values else ""
 
     @property
-    def expand(self) -> str:
-        """Get current $expand statement"""
+    def expand(self) -> expand_type:
+        """Get current $expand statement."""
         return self._expand
 
-    def _nested_commands(self, name: expand_keys, values: expand_values) -> str:
-
-        assert name in (
-            "expand",
-            "select",
-            "top",
-            "filter",
-            "orderby",
-        ), f"'{name}' is not a valid command!"
-
-        if name == "select":
-            return "$select=" + ",".join([key for key in values])
-
-        elif name == "filter":
-            return "$filter=" + " and ".join([key.strip() for key in values])
-
-        elif name == "orderby":
-            return f"$orderby=" + ",".join([f"{key} {order}" for key, order in values.items()])
-
-        elif name == "top":
-            return f"$top={values}"
-
-        else:  # nested expand
-            return self._recursive_expand(values)
-
-    def _recursive_expand(self, items: Dict[str, Optional[expand_subcommands]]) -> str:
-        return "$expand=" + ",".join(
-            [
-                f"{key}(" + ";".join([self._nested_commands(name, values) for name, values in value.items()]) + f")"
-                if value
-                else f"{key}"
-                for key, value in items.items()
-            ]
-        )
-
     @expand.setter
-    def expand(self, items: Dict[str, Optional[expand_subcommands]]):
+    def expand(self, items: expand_type):
         """Set $expand statement, with possible nested statements.
         Controls what data from related entities is returned.
 
-        Keys in 'items' indicate entities to expand.
-        Values in 'items' set to None expands to everything *immidiately* under that entity.
-
-        Otherwise, values in 'items' is a dict, which contains the statements inside the expand statement.
-        Valid keys for these are 'select', 'filter', 'top', 'orderby', and 'expand'.
-        Values under these keys should be constructed in the same manner as they are
-        when outside the expand statement, e.g. 'select' takes a List[str], 'top' an int, etc.
-
-        A nested expand statement takes another similar dict to 'items',
-        but it has some limitations (WEB API v9.1):
+        Nested expand statement limitations (WEB API v9.1):
 
         1. Nested expand statements can *only* be applied to **many-to-one/single-valued** relationships.
         This means nested expands for collections do not work!
@@ -518,23 +550,67 @@ class DynamicsClient:
         2. Each request can include a maximum of 10 expand statements.
         This applies to non-nested statements as well! There is no limit on the depth of nested
         expand statements, so long as the total is 10.
+
+        :param items: What linked tables (a.k.a. naviagation properties) to expand and
+                      what statements to apply inside the expanded tables.
+                      If items-dict value is set to an empty dict, no query options are used.
+                      Otherwise, valid keys for the items-dict are 'select', 'filter', 'top', 'orderby', and 'expand'.
+                      Values under these keys should be constructed in the same manner as they are
+                      when outside the expand statement, e.g. 'select' takes a List[str], 'top' an int, etc.
         """
 
-        self._expand = self._recursive_expand(items)
+        self._expand = items
+
+    def _compile_expand(self, items: expand_type = _sentinel) -> str:
+        if items is _sentinel:
+            items = self._expand
+
+        if not items:
+            return ""
+
+        return "$expand=" + ",".join(
+            [
+                f"{key}(" + ";".join([self._expand_commands(name, values) for name, values in value.items()]) + f")"
+                if value
+                else f"{key}"
+                for key, value in items.items()
+            ]
+        )
+
+    def _expand_commands(self, name: expand_keys, values: expand_values) -> str:
+        """Compile commands inside an expand statement."""
+
+        if name == "select":
+            return self._compile_select(values)
+        elif name == "filter":
+            return self._compile_filter(values)
+        elif name == "orderby":
+            return self._compile_orderby(values)
+        elif name == "top":
+            return self._compile_top(values)
+        elif name == "expand":
+            return self._compile_expand(values)
+        else:
+            raise KeyError(f"'{name}' is not a valid command!")
 
     @property
-    def filter(self) -> str:
+    def filter(self) -> filter_type:
         """Get current $filter statement"""
         return self._filter
 
     @filter.setter
-    def filter(self, items: Union[Set[str], List[str]]):
+    def filter(self, items: filter_type):
         """Set $filter statement. Sets the criteria for which entities will be returned.
 
-        Input a list-object to 'and' the items. Input a set-object to 'or' the items.
+        It is recommended to read the API Query Function Reference:
+        https://docs.microsoft.com/en-us/dynamics365/customer-engagement/web-api/queryfunctions
 
-        Below is a list of the standard operators. ":func:`dynamics.ftr`" object can be used to create these,
-        and a bunch of special operators. All filter conditions for string values are case insensitive.
+        ...and how to Query data using the Web API:
+        https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/query-data-web-api.
+
+        You can input the filters as strings, or use the included `ftr` object to construct them.
+
+        Below is a list of the standard operators:
 
         - **eq** = Equal: 'foo *eq* bar'
         - **ne** = Not Equal: 'foo *ne* bar'
@@ -549,39 +625,74 @@ class DynamicsClient:
         - **contains(key,'value')** = Key contains string: '*contains(foo,'bar')*'
         - **endswith(key,'value')** = Key ends with string: '*endswith(foo,'bar')*'
         - **startswith(key,'value')** = Key starts with string: '*startswith(foo,'bar')*'
+
+        :param items: If a list-object, 'and' the items. If a set-object, 'or' the items.
         """
 
-        if isinstance(items, set):
-            self._filter = "$filter=" + " or ".join([key.strip() for key in items])
-        elif isinstance(items, list):
-            self._filter = "$filter=" + " and ".join([key.strip() for key in items])
-        else:
+        if not isinstance(items, (set, list)):
             raise TypeError("Filter items must be either a set or a list.")
+        elif not items:
+            raise ValueError("Filter list cannot be empty.")
+
+        self._filter = items
+
+    def _compile_filter(self, values: filter_type = _sentinel):
+        if values is _sentinel:
+            values = self._filter
+
+        if not values:
+            return ""
+
+        if isinstance(values, set):
+            return "$filter=" + " or ".join([key.strip() for key in values])
+        elif isinstance(values, list):
+            return "$filter=" + " and ".join([key.strip() for key in values])
 
     @property
-    def top(self) -> str:
+    def top(self) -> int:
         """Get current $top statement"""
         return self._top
 
     @top.setter
     def top(self, number: int):
         """Set $top statement. Limits the number of results returned.
-        Note: You should not use 'top' with 'count'.
+        Note: You should not use 'top' and 'count' in the same query.
         """
-        self._top = f"$top={number}"
+        self._top = number
+
+    def _compile_top(self, number: int = _sentinel) -> str:
+        if number is _sentinel:
+            number = self._top
+
+        return f"$top={number}" if number != 0 else ""
 
     @property
-    def orderby(self) -> str:
+    def orderby(self) -> orderby_type:
         """Get current $orderby statement"""
         return self._orderby
 
     @orderby.setter
     def orderby(self, items: orderby_type):
         """Set $orderby statement. Specifies the order in which items are returned."""
-        self._orderby = f"$orderby=" + ",".join([f"{key} {order}" for key, order in items.items()])
+
+        if isinstance(items, dict):
+            raise TypeError("Orderby items must be a dict.")
+        elif not items:
+            raise ValueError("Orderby dict must not be empty.")
+
+        self._orderby = items
+
+    def _compile_orderby(self, values: orderby_type = _sentinel) -> str:
+        if values is _sentinel:
+            values = self._orderby
+
+        if not values:
+            return ""
+
+        return f"$orderby=" + ",".join([f"{key} {order}" for key, order in values.items()])
 
     @property
-    def count(self) -> str:
+    def count(self) -> bool:
         """Get current $count statement"""
         return self._count
 
@@ -589,9 +700,31 @@ class DynamicsClient:
     def count(self, value: bool):
         """Set $count statement. Include the count of entities that match the filter criteria in the results.
         Count will be the first item in the list of results.
-        Note: You should not use 'count' with 'top'.
+        Note: You should not use 'count' and 'top' in the same query.
         """
-        self._count = f"$count=true" if value else ""
+        self._count = value
+
+    def _compile_count(self, value: bool = _sentinel) -> str:
+        if value is _sentinel:
+            value = self._count
+
+        return f"$count=true" if value else ""
+
+    @property
+    def pagesize(self) -> int:
+        """Return currently set pagesize."""
+        return self._pagesize
+
+    @pagesize.setter
+    def pagesize(self, value: int):
+        """Specify the number of tables to return in a page. By default, this is set to 1000. Maximum is 5000."""
+
+        if value < 1:
+            ValueError(f"Value must be bigger than 0. Got {value}.")
+        elif value > 5000:
+            ValueError(f"Max pagesize is 5000. Got {value}.")
+
+        self._pagesize = value
 
     # TODO: apply-statement
     # TODO: Any and all statements
